#Version 0.2 22MAR2011
#History of revisions:
#1. Fixed from v0.1 (02MAY2011) Lattice plot of cluster centroids in function LF.cluster. The graph displayed a "panel function error" in addition to centroids.
#2. Changed from v0.1 (02MAY2011) output file name of kde bin cooridinates to eval_point.csv to match Jeff's ArcMap documentation. The output file name was grid.csv.
#Known issues:
#1. plot for 1-dimensional data in LF.kde does not display "contours"; use cbind(x,x) to turn 1D into 2D?

######################################################
### LOAD NECESSARY R PACKAGES
######################################################
library(ks)
library(e1071)
library(lattice)
library(clusterSim)
library(nnclust)
#library(akima); library(RColorBrewer); library(foreign)
######################################################
######################################################




######################################################
# FUNCTIONS FOR KERNEL DENSITY ESTIMATION AND IDENTIFICATION OF OUTLIERS
######################################################
LF.kde <- function(x, gridsize){ ### PERFORM kernel density estimation on 1-6 dimensional data
	if(ncol(x) > 6) { ### ERROR MESSAGE FOR DIMENSIONALITY OF DATA OBJECT
		print(noquote("ERROR: x must have fewer than 6 dimensions"))
		return() #Abort process		
	}
	
	if(ncol(x)>1){ #For >1-D Data
		Hpi <- Hpi.diag(x=x, pilot = "samse", binned = TRUE, bgridsize=rep(gridsize,ncol(x))) ### ESTIMATE DIAGONAL BANDWIDTH MATRIX for > 1D
		kde.binned <-kde(x, H = Hpi, binned = TRUE, bgridsize=rep(gridsize,ncol(x))) ### COMPUTE KERNEL DENSITY
		if(ncol(x)==2){
			plot(x, col=rgb(102,139,139,50,maxColorValue=255), pch = 16, cex = 0.25)
			plot(kde.binned, add=TRUE, cont=c(seq(0,70,10), seq(70,100,5)))
		}
		if(ncol(x)==3){
			plot(kde.binned, drawpoints=TRUE, ptcol=rgb(102,139,139,50,maxColorValue=255), cont=c(25,50,75,90,100), pch = 16, size = 0.25)
		}
	}

	if(ncol(x)==1){ #For 1-D data
		lab <- names(x)
		y <- x[,1]
		Hpi <- hpi(x=as.matrix(y), binned = TRUE, bgridsize=rep(gridsize,1)) ### ESTIMATE BANDWIDTH MATRIX for 1D
		kde.binned <-kde(y, H=Hpi, binned = TRUE, bgridsize=rep(gridsize,1)) ### COMPUTE KERNEL DENSITY
		kde.binned$names <- lab
		plot(kde.binned, drawpoints=TRUE, ptcol=rgb(102,139,139,50,maxColorValue=255), xlab=lab)
	}
	kde.binned$width <- LF.binwidth(kde.binned) ### IDENTIFY HALF-WIDTH OF BINS USED IN KDE
	kde.binned
}

LF.binwidth <- function(x){	
	if(is.numeric(x$x)) pts <- matrix(unlist(x$eval.points), ncol=1, byrow=FALSE) #Convert list of eval.points to matrix
	if(!is.numeric(x$x)) pts <- matrix(unlist(x$eval.points), ncol=length(x$eval.points), byrow=FALSE) #Convert list of eval.points to matrix
	width <- matrix(NA, ncol=ncol(pts))
	for(i in 1:(ncol(pts))){
		d <- mean(nnfind(pts[,i])$dist)/2
		width[1,i] <- d
	}	
	colnames(width) <- x$names
	write.csv(width, "bin_width.csv", row.names=FALSE)
	width
}

LF.outlier <- function(x, threshold=90) { ### MARK INDIVIDUAL DATA POINTS AS OUTLIERS BASED ON THRESHOLD
	lev <- contourLevels(x, prob=((100-threshold)/100)) ### DENSITY THRESHOLD FOR DEFINING OUTLIERS
	
	if(is.numeric(x$x)) pts <- matrix(unlist(x$eval.points), ncol=1, byrow=FALSE) #Convert list of eval.points to matrix
	if(!is.numeric(x$x)) pts <- matrix(unlist(x$eval.points), ncol=length(x$eval.points), byrow=FALSE) #Convert list of eval.points to matrix
	dims <- ncol(pts) #Dimension of KDE
	size <- nrow(pts) #Gridsize
	bins <- matrix(c(combn(rep(1:size,dims),dims)), ncol=dims, byrow=TRUE)
	bins <- unique(bins) #Indexed location of bins according to values in x$eval.points

	grid <- matrix(NA, nrow=nrow(bins), ncol=ncol(bins))
	for(i in 1:ncol(bins)){ #Loop to identify location of bins
		grid[,i] <- pts[bins[,i],i]
	}
	nn <- nnfind(grid, as.matrix(x$x))$neighbour ### FIND NEAREST BIN CENTROID TO EACH DATA POINT

	density <- numeric(nrow(bins))
	for(i in 1:nrow(bins)){ #Identify KDE of each bin
		if(ncol(bins)==1) density[i] <- x$estimate[bins[i]]
		if(ncol(bins)==2) density[i] <- x$estimate[bins[i,1], bins[i,2]]
		if(ncol(bins)==3) density[i] <- x$estimate[bins[i,1], bins[i,2], bins[i,3]]
		if(ncol(bins)==4) density[i] <- x$estimate[bins[i,1], bins[i,2], bins[i,3], bins[i,4]]
		if(ncol(bins)==5) density[i] <- x$estimate[bins[i,1], bins[i,2], bins[i,3], bins[i,4], bins[i,5]]
		if(ncol(bins)==6) density[i] <- x$estimate[bins[i,1], bins[i,2], bins[i,3], bins[i,4], bins[i,5], bins[i,6]]
	}

	index <- which(density>lev) #Identify which bins are non-outliers
	bins <- as.matrix(bins[index,]) #Indexed location of non-outlier bins according to values in x$eval.points
	grid <- matrix(NA, nrow=nrow(bins), ncol=ncol(bins)) #Receptacle for non-outlier bin locations
	for(i in 1:ncol(bins)){ #Loop to identify location of non-outlier bins
		grid[,i] <- pts[bins[,i],i]
	}
	colnames(grid) <- x$names
	write.csv(grid, "eval_points.csv", row.names=FALSE) #Save location of non-outlier bins

	density <- density[nn] #KDE for each object
	outlier <- ifelse(density > lev, 0, 1) #Identify individual objects as outliers/non-outliers
	list(outlier=outlier, density=density)
}

######################################################
######################################################



######################################################
# IDENTIFY OPTIMAL NUMBER OF CLUSTERS IN A DATA SET
######################################################
LF.cluster <- function(x, nclust=c(2:7), niter=30, psi=1.5, max=1000000){
	### STANDARDIZE DATA WRT NON-OUTLIERS INSIDE WILDLAND BLOCKS
	stand <- apply(x, 2, function(x) rbind(mean(x), sd(x))) ### MEAN AND SD OF VARIABLES
	for(i in 1:ncol(x)){
		x[,i] <- (x[,i]-stand[1,i])/stand[2,i]
	}

	for(c in nclust){
		print(noquote(paste("Number of clusters:",c, "         Time started:",(Sys.time()))))
	
		for(z in 1:niter){
			if(nrow(x) > max){ ### SAMPLE DATA FOR CLUSTER ANLAYSIS
				index <- sample(x, max) ### TAKE SRS WOR OF n = max FROM data
			}else
			if(nrow(x) <= max){ ### USE ALL DATA FOR CLUSTER ANALYSIS
				index <- 1:nrow(x)
			}
			
			### CLUSTER ANALYSIS
			fcm <- cmeans(x[index,], c, iter.max = 1000, dist = "euclidean", method = "cmeans", m=psi) # PERFROM CLUSTER ANALYSIS ON SAMPLED DATA
				
			### CATALOGUE CENTROIDS IN EACH ITERATION
			if(c==min(nclust) & z==1){
				centroids <- as.data.frame(cbind(c=c,iteration=z, fcm$centers[order(fcm$centers[,1]),])) ### CENTROIDS
			}
			if(c>min(nclust) | z>1){
				centroids <- rbind(centroids, cbind(c, iteration=z, fcm$centers[order(fcm$centers[,1]),]))
			}

			### CALCULATE VALIDITY INDICES
			FS <- fuksug(x=x[index,], centroids=fcm$centers, memb=fcm$membership, psi=psi) ### FUKUYAMA-SUGENO INDEX
			AWCD <- awcd(x=x[index,], centroids=fcm$centers, memb=fcm$membership, psi=psi) ### AVERAGE WITHIN CLUSTER DISTANCE INDEX			
			XB <- xiebeni(x=x[index,], centroids=fcm$centers, memb=fcm$membership, psi=psi) ### XIE-BENI AND XIE-BENI* INDICES
			DB <- index.DB(x=x[index,], cl = fcm$cluster, centrotypes="centroids", p=2, q=1) ### DAVIES-BOULDIN INDEX
			PBM <- pbm(x=x[index,], centroids=fcm$centers, memb=fcm$membership, psi=psi) ### PBM INDEX
			CH <- index.G1(x=x[index,], cl = fcm$cluster, centrotypes="centroids") ### CALINSKI=HARABASZ INDEX
			FUZZY.SIL <- fuzzy.sil(x=x[index,], centroids=fcm$centers, memb=fcm$membership, cluster=fcm$cluster, psi=psi, alpha=2) ### FUZZY SILHOUETTE INDEX

			### CATALOGUE VALIDITY INDICES
			if(c==min(nclust) & z==1){
				validity <- as.data.frame(cbind(c, z, FS, AWCD, XB$xb, XB$xb.star, DB$DB, PBM, CH, FUZZY.SIL$fuzzy.sil))
				names(validity) <- c("c", "iteration", "FS", "AWCD", "XB", "XB_star", "DB", "PBM", "CH", "FSil")
			}
			if(c>min(nclust) | z>1){
				validity <- rbind(validity, c(c, z, FS, AWCD, XB$xb, XB$xb.star, DB$DB, PBM, CH, FUZZY.SIL$fuzzy.sil))
			}
		}
	}

for(i in 1:ncol(centroids[,-c(1:2)])){ ### UNSTANDARDIZE LOCATION OF CLUSTER CENTROIDS
		centroids[,i+2] <- (centroids[,i+2] * stand[2,i] + stand[1,i])
	}

### GENERATE PLOTS
reps <- ceiling(niter/6) ### NUMBER OF TIMES TO REPEATE COLORS AND LINE STYLES BELOW
my.colors <- rep(c("light blue", "red", "grey", "green", "pink", "blue", "black", "orange", "dark green", "purple", "brown"), reps) 
palette(my.colors)
linetype <- rep(1:6, reps)
max.n <- max(nclust) ### MAXIMUM NUMBER OF CLUSTERS

### PLOT CENTROIDS
clust.lab <- paste(centroids$c, "Clusters", sep=" ")
p1 <- xyplot(centroids[,4]~centroids[,3]|clust.lab, group = iteration, data = centroids, as.table=TRUE, ylab = names(x)[2], xlab = names(x)[1], 
		col = centroids$iteration, pch = "", type = c("p", "g"), panel = function(x, y, subscripts, groups, ...){
			panel.superpose(x, y, subscripts, groups, ...)
			panel.text(x, y, labels = centroids$iteration[subscripts], col = centroids$iteration[subscripts])}
	)

###PLOT VALIDITY INDICES
p2<- xyplot(PBM + CH + FSil + XB + XB_star + DB + FS + AWCD ~ c, groups = iteration, data = validity, type = c("l", "g"), 
	col = my.colors, lty = linetype, xlim = c(2,max.n), xlab = list("Number of clusters (c)", cex = 1.0), ylab = list("Index value", cex = 1.0), 
	layout = c(3,3), scales = list(x=list(relation = "same", at = c(2:max.n), tck = c(1,1), cex = 1.0, alternating = FALSE), 
	y = list(relation = "free", cex = 1.0, tck = c(1,1))), par.strip.text=list(cex=1.0), strip=strip.custom(factor.levels=expression("PBMF", 
	"Calinski-Harabasz", paste("Fuzzy Silhouette (", alpha, " = 2.0)", sep = ""), "Xie-Beni", "Xie-Beni*", "Davies-Bouldin", "Fukuyama-Sugeno",
	"Average Within-Cluster Distance")), as.table = FALSE, between = list(x = 0.5),
	key = list(space = "right", title = "Iteration", cex = 1, cex.title = 1.25, columns = 1, lines = TRUE, text = list(as.character(1:niter)), 
	col = my.colors[1:niter], lty = linetype[1:niter]), 
	panel = function(x,y,...) {     
      	panel.abline(v=c(3:(max.n-1)), col = "light grey")
		panel.xyplot(x,y,...)}
	)  

print(noquote(paste("Finally done:", Sys.time())))

#par(ask=TRUE)
print(p1)
dev.new()
print(p2)

list(validity=validity, centroids=centroids, stand=stand, psi=psi)
}


LF.export <- function(x, k, iter=1){ #Export cluster centroids of optimal partition
	for(i in 1:ncol(x$centroids[,-c(1:2)])){ ### UNSTANDARDIZE LOCATION OF CLUSTER CENTROIDS
		x$centroids[,i+2] <- (x$centroids[,i+2] - x$stand[1,i])/x$stand[2,i] 
	}
	write.csv(x$centroids[x$centroids$c==k & x$centroids$iteration==iter, -c(1:2)], "centroids.csv", row.names=FALSE)
	row.names(x$stand) <- c("mean", "sd")
	write.csv(x$stand, "params.csv")
	junk <- as.matrix(x$psi)
	colnames(junk) <- "psi"
	write.csv(junk, "psi.csv",row.names=FALSE)
}
######################################################
######################################################



######################################################
# FUNCTIONS FOR VALIDITY INDICES
######################################################
#LF.dist <- function(x=x, y=y){ ### SLOWER BUT CLEANER FUNCTION FOR CALCULATING DISTANCES ON N-DIMENSIONAL DATA
#	sq.diff <- (x - mapply(rep, y, nrow(x)))^2
#	d <- apply(sq.diff, 1, sum)
#	d <- sqrt(d)	
#	d
#}

LF.dist <- function(x, y){ ### CREATES AND APPLIES DISTANCE FUNCTION FOR N-DIMENSIONAL DATA
	if(is.vector(x)) x <- t(as.matrix(x))
	if(is.vector(y)) y <- t(as.matrix(y))
	fxn <- "sqrt((x[,1]-y[,1])^2"
	if(ncol(x)>1){
		for(i in 2:ncol(x)){
			fxn <- paste(fxn, "+(x[,",i,"]-y[,",i,"])^2", sep="")
		}	
	}	
	fxn <- paste(fxn,")", sep="")
	dist.fxn <- function(x,y) eval(parse(text=fxn))
	d <- as.numeric(dist.fxn(x, y))
	d
}


### FUKUYAMA AND SUGENO INDEX (Zahid et al. 1999)
fuksug <- function(x, centroids, memb, psi){ ### 
	### THIS CODE PRODUCES SAME RESULT AS SUM(memb*cdist)-SUM(memb*mdist), PER EQUATION (10) IN ZAHID ET AL. 1999
	rec <- numeric()
	for(i in 1:nrow(centroids)){
		cdist <- LF.dist(x, centroids[i,])^2 ### SQUARED DISTANCES BETWEEN OBJECTS AND CENTROID i FOR J_m, A MEASURE OF COMPACTNESS
		mdist <- LF.dist(apply(x, 2, mean), centroids[i,])^2 ### SQUARED DISTANCE BETWEEN MEAN OF OBJECTS AND CENTROID i FOR K_m, A MEASURE OF SEPERATION		
		wdist <- (memb[,i]^psi) * (cdist-mdist) ### MEMBERSHIP VALUE WEIGHTED BY DIFFERENCE BETWEEN COMPACTNESS AND SEPERATION MEASURES
		rec <- sum(rec, wdist) ### SUM OF WEIGHTED SQUARED DISTANCES
	}
	as.numeric(rec)
}	

### XIE-BENI AND XIE-BENI* INDICES (Zahid et al. 1999; Xie & Beni 1999; Celikyilmaz & Turksen 2008)
xiebeni <- function(x, centroids, memb, psi){ 
	### CALCULATE XIE-BENI INDEX
	rec <- numeric()
	for(i in 1:nrow(centroids)){
		cdist <- LF.dist(x, centroids[i,])^2 ### SQUARED DISTANCES BETWEEN OBJECTS AND CENTROID i FOR J_m, A MEASURE OF COMPACTNESS
		wdist <- (memb[,i]^psi) * (cdist) ### MEMBERSHIP VALUE WEIGHTED BY DIFFERENCE BETWEEN COMPACTNESS AND SEPERATION MEASURES
		rec <- c(rec, sum(wdist)) ### SUM OF WEIGHTED SQUARED DISTANCES
	}
	numerator <- sum(rec)
	cmin <- min(dist(centroids))^2 ## MINIMUM DISTANCE BETWEEN CENTROIDS SQUARED
	denominator <- nrow(x)*cmin
	xb <- numerator/denominator ### XIE-BENI INDEX

	### CALCULATE XIE-BENI* INDEX
	numerator <- max(rec)
	xb.star <- numerator/denominator
	list(xb=as.numeric(xb), xb.star=as.numeric(xb.star))	
}

### AVERAGE WITHIN-CLUSTER DISTANCE (Campello & Hruschka 2006)
awcd <- function(x, centroids, memb, psi){
	rec <- numeric()
	for(i in 1:nrow(centroids)){
		cdist <- LF.dist(x, centroids[i,])^2 ### SQUARED DISTANCES BETWEEN OBJECTS AND CENTROID i FOR J_m, A MEASURE OF COMPACTNESS
		wdist <- (memb[,i]^psi) * (cdist) ### MEMBERSHIP VALUE WEIGHTED BY DIFFERENCE BETWEEN COMPACTNESS AND SEPERATION MEASURES
		numerator <- sum(wdist)
		denominator <- sum(memb[,i]^psi)
		rec <- sum(rec, numerator/denominator) ### SUM OF WEIGHTED SQUARED DISTANCES
	}
	awcd <- 1/nrow(centroids)*rec
	as.numeric(awcd)
}
	
### PBM INDEX (Pakhira et al. 2004; Zhang et al. 2008)
pbm <- function(x, centroids, memb, psi){
	E_1 <<- sum(LF.dist(x, apply(x, 2, mean))) ### SUM OF DISTANCES BETWEEN OBJECTS AND CENTROID OF DATA
	J_m <- 0 
	for(i in 1:nrow(centroids)){
		cdist <- LF.dist(x, centroids[i,]) ### SQUARED DISTANCES BETWEEN OBJECTS AND CENTROID i FOR J_m, A MEASURE OF COMPACTNESS
		wdist <- (memb[,i]^psi) * (cdist) ### MEMBERSHIP VALUE WEIGHTED BY DIFFERENCE BETWEEN COMPACTNESS AND SEPERATION MEASURES
		J_m <- J_m + sum(wdist) ### SUM OF WEIGHTED SQUARED DISTANCES
	}
	D_c <- max(dist(centroids))
	pbm <- ((1/nrow(centroids))*(E_1/J_m)*D_c)^2
	as.numeric(pbm)
}

### FUZZY SILHOUETTE INDEX (Campello & Hruschka 2006; Hruschka et al. 2006)
fuzzy.sil <- function(x, centroids, memb, cluster, psi, alpha){
	#a_i <- LF.dist(x, centroids[cluster,]) ### QUICK WAY TO CALCULATE DISTANCE BETWEEN EACH OBJECT AND ITS CORRESPONDING CLUSTER CENTROID
	d <- matrix(NA, nrow=nrow(x), ncol=nrow(centroids)) ### RECEPTACLE FOR DISTANCES BETWEEN EACH OBJECT AND ALL CENTROIDS
	for(i in 1:nrow(centroids)){
		d[,i] <- LF.dist(x, centroids[i,]) ### DISTANCE BETWEEN EACH OBJECT AND CENTROID i
	}

	s_i <- numeric(nrow(x)) ### RECEPTACLE FOR SILHOUETTES OF EACH OBJECT
	for(i in 1:nrow(centroids)){
		indx <- which(cluster==i)
		a_i <- d[indx, i] ### DISTANCE BETWEEN EACH OBJECT AND ITS CORRESPONDING CLUSTER CENTROID
		if(nrow(centroids)==2) b_i <- d[indx,-i] ### DISTANCE BETWEEN EACH OBJECT AND ITS NEXT CLOSEST CLUSTER CENTROID
		if(nrow(centroids)> 2)	b_i <- apply(d[indx,-i], 1, min) ### DISTANCE BETWEEN EACH OBJECT AND ITS NEXT CLOSEST CLUSTER CENTROID
		s_i[indx] <- (b_i - a_i)/apply(cbind(b_i, a_i), 1, max) ### CALCULATE SILHOUETTE FOR OBJECT i
	}

	d <- numeric(nrow(x)) ### RECEPTACLE FOR DIFFERENCES BETWEEN HIGHEST MEMBERSHIP AND SECOND HIGHEST MEMBERSHIP
	max.memb <- apply(memb, 1, max) ### HIGHEST MEMBERSHIP VALUE FOR EACH OBJECT
	for(i in 1:nrow(centroids)){
		indx <- which(cluster==i)
		if(nrow(centroids)==2) sec.memb <- memb[indx,-i] ### SECOND HIGHEST MEMBERSHIP FOR OBJECTS
		if(nrow(centroids)>2) sec.memb <- apply(memb[indx,-i], 1, max) ### SECOND HIGHEST MEMBERSHIP FOR OBJECTS
		d[indx] <- max.memb[indx] - sec.memb ### DIFFERENCE BETWEEN HIGHEST AND SECOND HIGHEST MEMBERSHIP VALUES
	}		
	fuzzy.sil <- sum(d^alpha * s_i)/sum(d^alpha) ### CALCULATION OF FUZZY SILHOUETTE CRITERION
	list(fuzzy.sil=as.numeric(fuzzy.sil), alpha=alpha)
}
######################################################
######################################################



######################################################
### JUNK? 
######################################################
#LF.out <- function(kde = kde, threshold = 90) { ###IDENTIFY OUTLIERS BASED ON DISTANCE OF EACH DATA POINT TO NON-OUTLIER BIN CENTROIDS
#	pts <- LF.bins(kde)
#	for(i in 1:ncol(kde$x)){ ### SCALE BIN CENTROIDS IN KDE BETWEEN [-1,1]
#		min <- min(kde$x[i])
#		max <- max(kde$x[i])
#		pts[,i] <- (pts[,i]-((min+max)/2))/((max-min)/2)
#	}
#
#	lev <- contourLevels(kde, prob=((100-threshold)/100)) ### THRESHOLD IN KDE DEFINING OUTLIERS
#
#	evalmat <- as.matrix(pts[pts$density>lev,-ncol(pts)]) ### BIN CENTROIDS WITH KDE > THESHOLD
#	stand <- apply(kde$x, 2, function(x) (x -((min(x)+max(x))/2))/((max(x)-min(x))/2)) ### SCALE DATA BETWEEN [-1,1]
#	interp <- nnfind(evalmat, stand) ### FIND DISTANCE OF INDIVIDUAL DATA POINTS TO NEAREST BIN CENTROID
#	
#	unique <- apply(pts[,-ncol(pts)], 2, unique) ### UNIQUE BIN CENTROIDS
#	bin <- apply(unique, 2, dist) ### DISTANCE BETWEEN UNIQUE BIN CENTROIDS
#	bin <- apply(bin, 2, min)/2 ### HALF THE DISTANCE BETWEEN ADJACENT BIN CENTROIDS
#	max.dist <- sqrt(sum(bin^2)) ### HALF THE DISTANCE BETWEEN DIAGONALLY ADJACENT BIN CENTROIDS
#	outlier <- ifelse(interp$dist<=max.dist, 0, 1) ### ALL CELLS WITHIN max.dist OF A BIN CENTROID NOT AN OUTLIER 
#
#	pts <- LF.bins(kde)
#	list(outlier=outlier, pts=pts)
#}
######################################################
######################################################
